# coding:utf-8
import select
import threading
import subprocess

import sys
import socket
import time
import pty
import os
import random
import paho.mqtt.client as mqtt

from .pkt_common import get_payload2, gen_pkt2
from .mqtt_common import start_mqtt_connection


class ClientTerm:
    def __init__(self):
        self.token = ''
        self.bridge_server = ''
        self.bridge_port = 1883
        self.bridge_topic_prefix = ''
        self.pty_master_fd = None
        self.pty_name = None
        self.topic_from_server = ''
        self.topic_from_client = ''

    def set_token(self, token):
        self.token = token

    def set_bridge(self, bridge_server, bridge_port, bridge_topic_prefix):
        self.bridge_server = bridge_server
        self.bridge_port = bridge_port
        self.bridge_topic_prefix = bridge_topic_prefix

    def start_pty(self):
        # open pseudo-terminal to interact with subprocess
        self.pty_master_fd, slave_fd = pty.openpty()

        # use os.setsid() make it run in a new process group, or bash job control will not be enabled
        self.proc = subprocess.Popen('bash',
                                     preexec_fn=os.setsid,
                                     stdin=slave_fd,
                                     stdout=slave_fd,
                                     stderr=slave_fd,
                                     universal_newlines=True
                                     )

        self.pty_name = os.ttyname(slave_fd)

    def close_pty(self):
        if self.pty_master_fd is not None:
            pass
        os.write(self.pty_master_fd, b'exit\n')
        os.close(self.pty_master_fd)
        self.pty_master_fd = None
        self.pty_name = None

    def on_connect(self, client, userdata, flags, rc):
        print('bridge connected')
        self.client.subscribe(self.topic_from_server)

    def on_message(self, client, userdata, msg):
        self.remote_data_to_local(msg.payload, self.pty_master_fd)

    def on_disconnect(self, client, userdata, rc):
        print('client disconnect:', client)

    def remote_data_to_local(self, remote_data, master_fd):
        output = remote_data
        command = get_payload2(output, self.token)
        if len(command) == 0:
            return
        os.write(master_fd, command)

    def local_data_to_remote(self):
        try:
            while self.proc.poll() == None:
                reta = select.select([self.pty_master_fd], [], [])
                if reta[0]:
                    output = os.read(self.pty_master_fd, 1024)
                    os.write(sys.stdout.fileno(), output)
                    # sys.stdout.write(output.decode('utf-8'))
                    sys.stdout.flush()

                    if len(output) > 0:
                        cmd = gen_pkt2(output, self.token)
                        if self.client.is_connected():
                            self.client.publish(self.topic_from_client, cmd)  # .decode('utf-8')
                        else:
                            print('local_data_to_remote: no connect. drop data')
        except Exception as e:
            print('local_data_to_remote error:')
            print(e)

        print('exit:local_data_to_remote')

        cmd = gen_pkt2(b'exited', self.token)
        self.client.publish(self.topic_from_client, cmd)  # .decode('utf-8')
        time.sleep(2)
        self.client.disconnect()
        self.close_pty()

    def start_server(self):
        self.start_pty()

        self.topic_from_client = self.bridge_topic_prefix + '/' + self.token + '/from_client'
        self.topic_from_server = self.bridge_topic_prefix + '/' + self.token + '/from_server'

        self.client = start_mqtt_connection(self.bridge_server, self.bridge_port, self)
        self.client.publish(self.topic_from_client,
                            gen_pkt2(('Hi Im connected :)' + self.pty_name + '\n').encode('utf-8'), self.token))

        self.local_data_to_remote()

        print('exit:client_one')
        sys.exit(0)



#!/usr/bin/env python
#coding:utf-8

import sys, getopt
import random
from .server_term import ServerTerm
from .client_term import ClientTerm


def main():
    try:
        opts, args = getopt.getopt(sys.argv[1:], "hst:b:p:f:", ['help', 'server', 'token', 'bridge', 'port', 'prefix'])
    except getopt.GetoptError:
        print('prog -h')

    join_a_client = False
    token = str(random.randrange(100000, 999999, 1))
    server = 'test.mosquitto.org'
    port = 1883
    prefix = ''
    for opt, arg in opts:
        if opt == ('-h', '--help'):
            print(
                'prog -s i am a server, connect to a client(no new session), . -t token, this is optional for client, must be set for server. -b bridge(default is test.mosquitto.org). -p port(default is 1883). -f topic prefix(default \'\') ')
            sys.exit()
        elif opt in ("-s", "--server"):
            join_a_client = True
        elif opt in ("-t", "--token"):
            token = arg
        elif opt in ("-b", "--bridge"):
            server = arg
        elif opt in ("-p", "--port"):
            port = int(arg)
        elif opt in ("-f", "--prefix"):
            prefix = prefix

    if join_a_client:
        print('starting a server to control remote client...')
        servert = ServerTerm()
        servert.set_token(token)
        servert.set_bridge(server, port, prefix)
        servert.start_server()
    else:
        print('starting a client to be controlled...')
        client = ClientTerm()
        client.set_token(token)
        client.set_bridge(server, port, prefix)
        if len(prefix) > 0:
            print('client started, use: term_from_nat -s -t  ' + token + ' -b ' + server + ' -p ' + str(
                port) + ' -f ' + prefix + '\n to start the server')
        else:
            print('client started, use: term_from_nat -s -t  ' + token + ' -b ' + server + ' -p ' + str(
                port) + '\n to start the server')
        client.start_server()


if __name__ == '__main__':
    main()


import select
import threading
import subprocess

import sys
import socket
import time
import pty
import os
import random
from pkt_common import get_payload2, gen_pkt2

g_is_exit = 0

g_tk=str(random.randrange(100000,999999,1))

def remote_data_to_local(client, master_fd):
    global g_is_exit
    while True:
        if g_is_exit == 1:
            print('exit:remote_data_to_local-1')
            sys.exit(0)

        reta = select.select([client.fileno()], [], [client.fileno()], 0.5)
        if reta[0]:
            output = os.read(client.fileno(), 8192)
            command = get_payload2(output,g_tk)

            if len(command) == 0:
                continue

            os.write(master_fd, command)
        elif reta[2]:
            # exception?
            print('exit:remote_data_to_local-3')
            sys.exit(0)


def local_data_to_remote(proc, master_fd, client):
    try:
        while proc.poll() == None:
            reta = select.select([master_fd], [], [])
            if reta[0]:
                output = os.read(master_fd, 1024)
                os.write(sys.stdout.fileno(), output)
                #sys.stdout.write(output.decode('utf-8'))
                sys.stdout.flush()

                if len(output) > 0:
                    cmd = gen_pkt2(output,g_tk)
                    client.send(cmd)  # .decode('utf-8')
    except Exception as e:
        print('local_data_to_remote error:')
        print(e)

    print('exit:local_data_to_remote')
    global g_is_exit
    g_is_exit = 1
    time.sleep(2)
    client.close()
    os.close(master_fd)
    sys.exit(0)


def client_one(remote_host, remote_port):
    client = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    client.connect((remote_host, remote_port))
    print('client connected to:', remote_host, remote_port)

    # open pseudo-terminal to interact with subprocess
    master_fd, slave_fd = pty.openpty()

    # use os.setsid() make it run in a new process group, or bash job control will not be enabled
    proc = subprocess.Popen('bash',
                            preexec_fn=os.setsid,
                            stdin=slave_fd,
                            stdout=slave_fd,
                            stderr=slave_fd,
                            universal_newlines=True
                            )

    tty_name = os.ttyname(slave_fd)
    client.send( gen_pkt2( ('Hi Im connected :)' + tty_name + '\n').encode('utf-8'), g_tk))

    t = threading.Thread(target=remote_data_to_local, args=(client, master_fd))
    t.start()

    local_data_to_remote(proc, master_fd, client)

    print('exit:client_one')
    client.close()
    sys.exit(0)


if __name__ == "__main__":
    if len(sys.argv) < 3:
        print('usage: prog remote_host remote_port')
        sys.exit(0)
    remote_host = sys.argv[1]
    remote_port = int(sys.argv[2])
    if len(sys.argv)>3:
        g_tk = str(sys.argv[3])
    client_one(remote_host, remote_port)



# coding:utf-8
import socket
import os
import select
import termios
import threading
import tty
import atexit
import fcntl
import struct
import sys
import time
import termios
from pkt_common import get_payload2, gen_pkt2
import random

g_old_settings = None
g_is_exit = 0
g_tk = str(random.randrange(100000, 999999, 1))


def tty_to_raw():
    fd = sys.stdin.fileno()
    tty.setraw(fd)

def tty_setting_read():
    global g_old_settings
    fd = sys.stdin.fileno()
    g_old_settings = termios.tcgetattr(fd)


@atexit.register
def tty_restore():
    global g_old_settings
    fd = sys.stdin.fileno()
    termios.tcsetattr(fd, termios.TCSADRAIN, g_old_settings)


def read_local_and_send_to_remote(client):
    global g_is_exit
    try:
        fd = sys.stdin.fileno()
        while 1:
            # command = input()
            command = os.read(fd, 32)
            # for icmd in command:
            client.send(gen_pkt2(command, g_tk))

    except Exception as e:
        print('[-] Caught exception: ' + str(e))
        try:
            client.close()
        except:
            pass
        print('exit:read_local_and_send_to_remote-1')
        sys.exit(1)


def recv_remote_and_display_to_local(client):
    global g_is_exit
    while True:
        if g_is_exit:
            print('exit: recv_remote_and_display_to_local')
            sys.exit(0)
        try:
            reta = select.select([client.fileno()], [], [client.fileno()], 0.5)
            if reta[0]:
                output = os.read(client.fileno(), 8192)
                output = get_payload2(output, g_tk)
                if len(output)>0:
                    os.write(sys.stdout.fileno(), output)
                    sys.stdout.flush()
            elif reta[2]:
                print('exit: recv_remote_and_display_to_local-2')
                sys.exit(0)

        except Exception as e:
            print(e)
            print('exit: recv_remote_and_display_to_local-e')
            sys.exit(0)


def server_one(server_port=12345):
    try:
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        sock.bind(('0.0.0.0', server_port))
        sock.listen(100)
        print('[+] Listening:', server_port)
        client, addr = sock.accept()
    except Exception as e:
        print('[-] Listen/Bind/Accept failed: ' + str(e))
        sys.exit(1)
    print('[+] Got a connection!')
    tty_setting_read()
    tty_to_raw()

    # recv remote and display
    t = threading.Thread(target=recv_remote_and_display_to_local, args=(client,))
    t.start()
    # read input and send
    read_local_and_send_to_remote(client)


if __name__ == '__main__':
    server_port = 12345
    if len(sys.argv) > 1:
        server_port = int(sys.argv[1])
    server_one(server_port)



# coding:utf-8

import socket
import os
import select
import termios
import threading
import tty
import atexit
import fcntl
import struct
import sys
import time
import termios
import random
from .pkt_common import get_payload2, gen_pkt2
from .mqtt_common import start_mqtt_connection


g_old_settings = None

def tty_to_raw():
    fd = sys.stdin.fileno()
    tty.setraw(fd)


def tty_setting_read():
    global g_old_settings
    fd = sys.stdin.fileno()
    g_old_settings = termios.tcgetattr(fd)


@atexit.register
def tty_restore():
    global g_old_settings
    if g_old_settings is not None:
        fd = sys.stdin.fileno()
        termios.tcsetattr(fd, termios.TCSADRAIN, g_old_settings)


class ServerTerm:
    def __init__(self):
        self.token = ''
        self.bridge_server = ''
        self.bridge_port = 1883
        self.bridge_topic_prefix = ''
        self.pty_master_fd = None
        self.pty_name = None
        self.topic_from_server = ''
        self.topic_from_client = ''
        self.is_exit=0

    def set_token(self, token):
        self.token = token

    def set_bridge(self, bridge_server, bridge_port, bridge_topic_prefix):
        self.bridge_server = bridge_server
        self.bridge_port = bridge_port
        self.bridge_topic_prefix = bridge_topic_prefix

    def local_data_to_remote(self):
        try:
            fd = sys.stdin.fileno()
            while 1:
                if self.is_exit:
                    break

                command = os.read(fd, 32)
                self.client.publish(self.topic_from_server, gen_pkt2(command, self.token))

        except Exception as e:
            print('[-] Caught exception: ' + str(e))
            try:
                self.client.disconnect()
            except:
                pass
            print('exit:read_local_and_send_to_remote-1')
            sys.exit(1)

    def remote_data_to_local(self, payload):
        output = payload
        output = get_payload2(output, self.token)
        if len(output) > 0:
            os.write(sys.stdout.fileno(), output)
            sys.stdout.flush()

        if output==b'exited':
            self.client.disconnect()
            self.is_exit=1

    def on_connect(self, client, userdata, flags, rc):
        self.client.subscribe(self.topic_from_client)
        tty_setting_read()
        tty_to_raw()
        print('connected')

    def on_message(self, client, userdata, msg):
        self.remote_data_to_local(msg.payload)

    def on_disconnect(self, client, userdata, rc):
        print('client disconnect:', client)

    def start_server(self):
        self.topic_from_client = self.bridge_topic_prefix + '/' + self.token + '/from_client'
        self.topic_from_server = self.bridge_topic_prefix + '/' + self.token + '/from_server'

        try:
            self.client = start_mqtt_connection(self.bridge_server, self.bridge_port, self)
        except Exception as e:
            print('[-] Listen/Bind/Accept failed: ' + str(e))
            sys.exit(1)

        self.local_data_to_remote()



# coding: utf-8

import setuptools

with open("README.md", "r", encoding='utf-8') as fh:
    long_description = fh.read()
	
setuptools.setup(
    name="term_from_nat",
    version="0.0.8",
    author="jiangshan00000",
    author_email="710806594@qq.com",
    description="A remote term connect from one nat to another nat",
	long_description=long_description,
	long_description_content_type = "text/markdown",
    url="https://github.com/Jiangshan00001/term_from_nat",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    install_requires=['paho-mqtt',],
    entry_points={
            'console_scripts': ['term_from_nat=term_from_nat.term_from_nat:main']
        },
    scripts=['term_from_nat_cli.py'],

)


from paho.mqtt import client as mqtt


def on_connect_default(client, userdata, flags, rc):
    userdata.on_connect(client, userdata, flags, rc)


def on_message_default(client, userdata, msg):
    userdata.on_message(client, userdata, msg)


def on_disconnect_default(client, userdata, rc):
    userdata.on_disconnect(client, userdata, rc)


def start_mqtt_connection(server, port, user_data, on_connect=None, on_message=None, on_disconnect=None):
    client = mqtt.Client(transport='tcp')
    client.user_data_set(user_data)
    client.on_connect = on_connect if on_connect is not None else on_connect_default
    client.on_message = on_message if on_connect is not None else on_message_default
    client.on_disconnect = on_disconnect if on_connect is not None else on_disconnect_default
    print('connecting:', server, port)
    client.connect(server, port, 60)
    client.loop_start()
    return client



# coding:utf-8
import base64
import json


def account_to_topic(account, password):
    return '/' + account + '/'



def gen_pkt(payload: str, tk: str) -> bytes:
    try:
        pkt = {'payload': payload, 'tk': tk, 'protocol': 'pty'}
        ret = bytes(json.dumps(pkt), 'utf-8')
        ret_len = len(ret)
        ret_len_str = '{0:08d}'.format(ret_len)
        ret = b'\xaa' + ret_len_str.encode('utf-8') + b'\xab' + ret
    except Exception as e:
        print('gen_pkt error:')
        print(e)
        print(payload)
    return ret

head_len_len=8
HEAD_SIGN=b'\xaa'
HEAD_SIGN2 = b'\xab'

def gen_pkt2(payload: bytes, tk: str) -> bytes:
    try:
        pkt = {'payload':  base64.b64encode(payload).decode('utf-8'), 'tk': tk, 'protocol': 'pty'}
        ret = bytes(json.dumps(pkt), 'utf-8')
        ret_len = len(ret)
        ret_len_str = '{0:08d}'.format(ret_len)
        ret = HEAD_SIGN + ret_len_str.encode('utf-8') + HEAD_SIGN2 + ret
    except Exception as e:
        print('gen_pkt2 error:')
        print(e)
        print(payload)
    return ret

def get_payload2(pkt: bytes, tk: str) -> bytes:
    head_len_len = 8
    try:
        if pkt[0] != 0xaa:
            return b''
        len_str = pkt[1:1+head_len_len].decode('utf-8')
        pkt_len = int(len_str)
        if len(pkt) < (pkt_len + head_len_len + 2):
            return b''
        if len(pkt) > (pkt_len + head_len_len + 2):
            pkt_str = pkt[head_len_len+2:pkt_len +head_len_len+2].decode('utf-8')
            pkt_obj = json.loads(pkt_str)
            pkt_append = get_payload2(pkt[pkt_len + head_len_len+2:], tk)
            return base64.b64decode(pkt_obj['payload']) + pkt_append
        else:
            pkt_str = pkt[head_len_len+2:].decode('utf-8')
            pkt_obj = json.loads(pkt_str)
            return base64.b64decode(pkt_obj['payload'])
    except Exception as e:
        print('get_payload2:error.', e, len(pkt))
        return b''



def get_payload(pkt: bytes, tk: str) -> str:
    try:
        if pkt[0] != 0xaa:
            return ''
        len_str = pkt[1:9].decode('utf-8')
        pkt_len = int(len_str)
        if len(pkt) < (pkt_len + 8 + 2):
            return ''
        if len(pkt) > (pkt_len + 8 + 2):
            pkt_str = pkt[10:pkt_len + 10].decode('utf-8')
            pkt_obj = json.loads(pkt_str)
            pkt_append = get_payload(pkt[pkt_len + 10:], tk)
            return pkt_obj['payload'] + pkt_append
        else:
            pkt_str = pkt[10:].decode('utf-8')
            pkt_obj = json.loads(pkt_str)
            return pkt_obj['payload']
    except Exception as e:
        print('get_payload:json decode error.', e, len(pkt))
        return ''
